/*
 * @(#)RunUnot.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */

import unot.*;
import java.io.*;
import java.util.*;
import gnu.getopt.*;

/**
 * RunUnot.java
 *
 * Created: Thu Aug  7 19:07:56 2003
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.11 $
 * @see unot.Unot
 * @see <a href="http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf">
 * Discovering Frequent Substructures in Large Unordered Trees (PDF)</a>
 */
public class RunUnot
{

	/** ѥνϼ³ڤؤ */
    static int PATTERN_WRITER_TYPE = PatternWriter.WRITER_DEPTH;

    /** إץå */
	public static final String OPTION_INFORM_STRING =
	" -f <format id>	: \n"+
	"  "+DocSet.READER_PARENT+": parent tree format (default);\n" +
	"  "+DocSet.READER_CHILD +": child  tree format;\n" +
	"  "+DocSet.READER_XML   +": xml format;\n"; 

	public static final String OPTION_OUTFORM_STRING =
	" -g <format id>	: \n"+
	"  "+PatternWriter.WRITER_DEPTH+": depth-label format (default); \t"+
	"  "+PatternWriter.WRITER_PARENT+": parent tree format; \n"+
	"  "+PatternWriter.WRITER_TREE+  ": character tree format; \t"+
	"  "+PatternWriter.WRITER_SERIAL+": serialized  format; \n"+
	" -gi                  \t set on the indentation in output format  \n"+
	"";

    public static final String USAGE_STRING =
	"Usage: java [Java OPTIONS] RunUnot [OPTIONS...] MINSUP FILE [FILES...] \n\n" +
	"ARGS \n"+
	" MINSUP: \t a minimum support threshold in percentage (0 < minsup <= 100)\n" +
	" FILE [FILES]:\t input tree files. For the format see -f option.\n" +
	"   A special file name \"-\" indicates the STDIN and the rest of the file\n"+
	"   arguments are ignored. \n"+ 
	"OPTIONS \n"+
	" -m NUM | --maxsize NUM \t set maximum pattern size. \n" +
	OPTION_INFORM_STRING +
	OPTION_OUTFORM_STRING+
	" -o | --out <outfile> \t set the output to <outfile>. System.out is default\n"+
	" -q | --quiet    \t no output. \n" +
	" -v | --verbose  \t some log info is output. \n" +
	" -vv             \t -v option + tree output. \n" +
	"EXPERIMENTAL\n" + 
	" -b | --non-bc   \t no BC (balance check) optimization technique \n" +
	" -e | --non-od   \t no OD (occurrence deliver) optimization tech. \n" +
	" -l | --rootcount\t print the root occurrence count with each pattern \n" +
	" -d | --doccount \t print the doc  occurrence count with each pattern \n" +
	" -L | --rootlist \t print the root occurrence list with each pattern \n" +
	" -D | --doclist  \t print the doc  occurrence list with each pattern \n" +
	" -t | --tag <label> \t mining only subtrees of the nodes with <label>.\n" +
	" -pd|               \t do not mine and only print a dataset \n" +
	"INFO\n" + 
	" -V | --version  \t print version number of this program. \n" +
	" -h | --help     \t print this help. \n";

    /** ¼ɬפο */
    public static final int ARG_AT_LEAST = 2;

    /** ѥ󥵥κ͡ǥեȤ̵硥 */
    public static int maxCodeSize = Unot.MAX_SIZE_NO_LIMIT;

    /** åΥ٥롥 */
    public static int logLevel = UnotLog.LOG_DEFAULT;

    /** Ǿݡ͡(0 < minSup <= 1)  Double ͤͿ롥 */
    public static double minSup = 1;

    /** ϥեΥꥹȡ */
    public static List inputFiles = new ArrayList();

    /** ڥޥ˥ Freqt Ȥʦ ver. */
    //public static boolean useFreqt = false;

    /** ̵ڡʽڡˤȤɸ=trueˡ */
    public static boolean useUnot = true;

    /** иʬ۵ˡODȤɸ=trueˡ */
    public static boolean useOD = true;

    /** Х󥹥åˡBCȤɸ=trueˡ */
    public static boolean useBC = true;

    /** ǡڤ°(AT)Ȥɸ=falseˡ */
    public static boolean useAT = false;

	/** ѥϤ򥤥ǥȤ */
	public static boolean usePatIndent = false;

	/** ƥȡϥեɤ߹ߡϤ */
	public static boolean doPrintDocSet = false;

	/** ƥȡʸΡɤ٥ǻꤹ롥Υ٥롥*/
	public static String rootDocLabel = null;

	/** ѥϼ³ڤؤ */
    static int DATA_READER_TYPE    = DocSet.READER_PARENT;

	/** ϥե */
	static String outFileName = null; 
	static PrintStream out = System.out; //default output stream

    /**
     * Gets argument values.
     *
     * @param args a <code>String[]</code> value
     */
    public static void readOptionsAndArgs(String[] args) {

        if (args.length == 0) {
            System.out.println(USAGE_STRING);
            System.exit(Unot.EXIT_SUCCESS);
        }

        int c;
        LongOpt[] longopts = {
            new LongOpt("non-od",     LongOpt.NO_ARGUMENT,       null, 'e'),
            new LongOpt("non-bc",     LongOpt.NO_ARGUMENT,       null, 'b'),
            new LongOpt("rootlist",   LongOpt.NO_ARGUMENT,       null, 'L'),
            new LongOpt("rootcount",   LongOpt.NO_ARGUMENT,      null, 'l'),
            new LongOpt("freqt",      LongOpt.NO_ARGUMENT,       null, 'f'),
            new LongOpt("maxsize",    LongOpt.REQUIRED_ARGUMENT, null, 'm'),
            new LongOpt("output",     LongOpt.NO_ARGUMENT,       null, 'o'),
            new LongOpt("quiet",      LongOpt.NO_ARGUMENT,       null, 'q'),
            new LongOpt("non-unot",   LongOpt.OPTIONAL_ARGUMENT, null, 'u'),
            new LongOpt("tag",   	  LongOpt.REQUIRED_ARGUMENT, null, 't'),
            new LongOpt("verbose",    LongOpt.OPTIONAL_ARGUMENT, null, 'v'),
            new LongOpt("version",    LongOpt.NO_ARGUMENT,       null, 'V'),
            new LongOpt("help",       LongOpt.OPTIONAL_ARGUMENT, null, 'h'),
        };
        Getopt g = new Getopt("RunUnot", args, "abdlLDf:g:p:m:o:qv::t:uVh::", longopts);

        try {

            while ((c = g.getopt()) != -1) {
                switch (c) {

                case 'a':
                    RunUnot.useAT = true;
                    break;

                case 'b':
                    RunUnot.useBC = false;
                    break;

                case 'e':
                    RunUnot.useOD = false;
                    break;

                case 'u':
                    RunUnot.useUnot = false;
                    break;

                case 'l':
                    Unot.useRootCount   = true;
                    break;

                case 'd':
                    Unot.useDocCount   = true;
                    break;

                case 'L':
                    Unot.useRootList   = true;
                    break;

                case 'D':
                    Unot.useDocList   = true;
                    break;

                case 'm':
                    RunUnot.maxCodeSize = Integer.parseInt(g.getOptarg());
                    break;

//                 case 'z':
//                     RunUnot.outFileName = new String(g.getOptarg());
//                     break;

                case 'o':
                    RunUnot.outFileName = new String(g.getOptarg());
                    break;

                case 'q':
                    RunUnot.logLevel = UnotLog.LOG_WARNING;
                    break;

                case 'f':
                    RunUnot.DATA_READER_TYPE
						= Integer.parseInt(g.getOptarg());
					UnotLog.log("@log -f infile format " +
								RunUnot.DATA_READER_TYPE + "\n");
                    break;

                case 'g':
                    if (g.getOptarg() == null) { // -g
						; //do nothing
                    } else if (g.getOptarg().equals("i")) { // -gi
						RunUnot.usePatIndent = true;
                    } else { // -g num 
						RunUnot.PATTERN_WRITER_TYPE
							= Integer.parseInt(g.getOptarg());
						UnotLog.log("@log -g outfile format " +
									RunUnot.PATTERN_WRITER_TYPE + "\n");
                    }
                    break;

                case 'p':
                    if (g.getOptarg() == null) { // -p
						; //do nothing
                    } else if (g.getOptarg().equals("d")) { // -pd
						//ѡǡνϤΤ
						RunUnot.doPrintDocSet = true;
                    } else { // -p other 
						; //do nothing
                    }
                    break;

                case 't':
                    if (g.getOptarg() == null) { // -a
						; //do nothing
                    } else {
						//ѡΥ٥Ϳʸ񽸹ꤷơޥ˥.
						RunUnot.rootDocLabel = g.getOptarg(); 
						UnotLog.log("@log -at: rootlabel "+RunUnot.rootDocLabel+"\n");
					}
                    break;

                case 'v':
                    if (g.getOptarg() == null) { // -v
                        RunUnot.logLevel = UnotLog.LOG_INFO;
                    } else if (g.getOptarg().equals("v")) { // -vv
                        RunUnot.logLevel = UnotLog.LOG_DEBUG;
                    } else {
                        throw new IllegalArgumentException("Error: Verbose flag is -v or -vv.");
                    }
                    break;

                case 'V':
                    System.out.println(Unot.VERSION_STRING);
                    System.exit(Unot.EXIT_SUCCESS);
                    break;

                case 'h':
                    System.out.println(USAGE_STRING);
                    System.exit(Unot.EXIT_SUCCESS);
                    break;

                case ':': /* NO BREAK */
                case '?': /* NO BREAK */
                default:
                    throw new IllegalArgumentException("Error: Unknown option: ["+((char)c)+"]");
                }
            }

            // Process the rest of arguments
            int optind = g.getOptind();

			System.out.println("debug: args.length="+args.length+
							   " getOptind()="+g.getOptind());

            // First Argument
            if (optind >= args.length) //no minsup
				throw new IllegalArgumentException("Too few arguments (minsup expected)");
            RunUnot.minSup = (Double.parseDouble(args[optind++]) / 100);

            // Second (and the following) Argument(s)
            for (; optind < args.length; optind++) {
                String filename = args[optind];
				if (filename.charAt(0)=='@') {
					//a special indicator. 
					RunUnot.inputFiles.add(filename);
				}
				else if (filename.equals("-")) {
					//the indicator for reading from the stdin
					RunUnot.inputFiles.add(filename); 					
					break; //skip the rest of arguments
				}
				else {
					//an ordinary filename: test if it is accessible
					if (!(new File(filename)).exists())
						throw new FileNotFoundException(filename);
					if (!(new File(filename)).canRead())
						throw new IOException();
					RunUnot.inputFiles.add(filename);
				}

            }
			if (RunUnot.inputFiles.size() <= 0) //no file read
				throw new IllegalArgumentException("Too few arguments (file names expected)");

        } catch (IllegalArgumentException e) {
            System.err.println("Error: IllegalArgument: "+e.getMessage());
			System.out.println(USAGE_STRING);
            e.printStackTrace();
            System.exit(Unot.EXIT_FAILURE);
        } catch (FileNotFoundException e) {
            System.err.println("Error: File not found: "+e.getMessage());
            e.printStackTrace();
            System.exit(Unot.EXIT_FAILURE);
        } catch (IOException e) {
            System.err.println("Error: File can't be read:");
            e.printStackTrace();
            System.exit(Unot.EXIT_FAILURE);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(Unot.EXIT_FAILURE);
        }

    } // getOptions

	static void setupPatternWriter(int type, DocSet docSet) {
		if (type == PatternWriter.WRITER_TREE) {
			UnotPattern.setPatternWriter(new TreePatternWriter(docSet));
		}
		else if (type == PatternWriter.WRITER_PARENT) {
			UnotPattern.setPatternWriter(new ParentPatternWriter(docSet));
		}
		else if (type == PatternWriter.WRITER_SERIAL) {
			UnotPattern.setPatternWriter(new SerializeUnotPatternWriter(docSet));
		}
		else if (type == PatternWriter.WRITER_DEPTH) {
		UnotPattern.setPatternWriter(new DepthPatternWriter(docSet));
		}
		else {
			System.out.println("error: RunUnot: -f option out of range: "+
							   type+"\n"+OPTION_OUTFORM_STRING);
			System.exit(1);
		}

		if (RunUnot.usePatIndent) {
			UnotPattern.getPatternWriter().setIndent(true);
		}
		else {
			UnotPattern.getPatternWriter().setIndent(false);
		}
	}

    /**
     * Starts main() routine.
     *
     * @param args a <code>String[]</code> value
     */
    public static void main(String[] args)
    {

        // getopt
        readOptionsAndArgs(args);

        // Loglevel
        UnotLog.setLogLevel(RunUnot.logLevel); // optional

        // DocSet
		DocSet docSet = new DocSet();
		if (RunUnot.useAT) {
			docSet = new AttrDocSet();
			
		}
		else {
			docSet = new DocSet();
		}

		//DocSet: ѥʲե饰Хǡʲ롥
		if (RunUnot.usePatIndent) 
			docSet.setIndent(true);

		//DocSet: UnotPattern: Setup the pattern writer
		setupPatternWriter(PATTERN_WRITER_TYPE, docSet);

		// DocSet: reading the input files
        docSet.readFiles(RunUnot.inputFiles);
        UnotLog.log("@log DocSet: " + docSet + "\n");

        // DocSet: °ڤ
		if (RunUnot.useAT) {
			((AttrDocSet) docSet).pruneSiblings();
		}

		// Prepare the output stream
		try {
			//outFileName
			if (RunUnot.outFileName != null) 
				out = new PrintStream(new FileOutputStream(RunUnot.outFileName));
		}
		catch(IOException e){
			System.out.println("main: cannot open outFile: "+outFileName);
			System.exit(1);
		}
		
        // Select a mining engine 
        Unot unot;
		if (RunUnot.useUnot) {
			if (RunUnot.useOD) {
				unot = new UnotOD(docSet); //Unot + OD
			} else {
				unot = new UnotNonOD(docSet);   //Unot - OD
			}
		} else {
			unot = new UnotFreqt(docSet); //Freqt - OD
		}


		unot.useBalanceCheck = RunUnot.useBC; 

        unot.setMaxCodeSize(RunUnot.maxCodeSize); // optional

		if (RunUnot.doPrintDocSet) {
			//Print docSet
			UnotLog.log("@log Printing the document tree set\n");
			docSet.writeFile(out);
		}
		else if (RunUnot.rootDocLabel != null) {
			// Compute Document roots
			int labelId = docSet.getLabelDict().lookupId(RunUnot.rootDocLabel);
			NodeList rootList = docSet.getANodeList(docSet.getDocList(), labelId);
			UnotLog.log("@log tag="+RunUnot.rootDocLabel+
						" numRootMatch="+rootList.size()+"\n");
			
			// Mine datasets
			long start, time;
			UnotLog.log("@log Start tree mining\n");
			start = System.currentTimeMillis();
			PatSet patSet = unot.run(rootList, RunUnot.minSup);
			time = System.currentTimeMillis() - start;

			// Print patSet
			patSet.writeFile(out);
			patSet.writeStatFile(System.out);
			UnotLog.log("@log Run_Time: " + time + " [ms]\n");
		}
		else {
			// Mine datasets
			long start, time;
			UnotLog.log("@log Start tree mining\n");
			start = System.currentTimeMillis();
			PatSet patSet = unot.run(RunUnot.minSup);
			time = System.currentTimeMillis() - start;

			// Print patSet
			patSet.writeFile(out);
			patSet.writeStatFile(System.out);
			UnotLog.log("@log Run_Time: " + time + " [ms]\n");
		}

    } // end of main()


} // RunUnot
