/*
 * @(#)XmlFormatter.java
 *
 * Copyright 2003, Shinji Fusanobu all rights reserved.
 */

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;

/**
 * XmlFormatter.java
 *
 * Compile & Execute:
 *
 * javac -classpath ".:lib/xercesImpl.jar" src/app/XmlFormatter.java
 * java -cp "classes:lib/xercesImpl.jar:lib/xmlParserAPIs.jar" XmlFormatter -s data/sample.xml
 *
 *
 * Created: Thu Jan  8 17:50:17 2004
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version 1.0
 */
public class XmlFormatter
{

    public static boolean shuffleSiblings = false;
    public static boolean purgeTextNodes = false;
    public static boolean purgeAttributes = false;
    public static int maxElementSize = 0;
    public static int elementSize = 0;

    /**
     * Describe <code>main</code> method here.
     *
     * @param argv a <code>String[]</code> value
     */
    public static void main(String[] argv)
    {
        if (argv.length == 0) {
            printUsage();
            return;
        }

        XmlFormatter formatter = new XmlFormatter();
        for (int i = 0; i < argv.length; i++) {
            String arg = argv[i];
            if (arg.startsWith("-")) {
                String option = arg.substring(1);
                if (option.equalsIgnoreCase("s")) {
                    formatter.shuffleSiblings = option.equals("s");
                    continue;
                }
                if (option.equalsIgnoreCase("t")) {
                    formatter.purgeTextNodes = option.equals("t");
                    continue;
                }
                if (option.equalsIgnoreCase("a")) {
                    formatter.purgeAttributes = option.equals("a");
                    continue;
                }
                if (option.startsWith("m")) {
                    formatter.maxElementSize = Integer.parseInt(option.substring(1));
                    continue;
                }
                if (option.equals("h") || option.equals("-help")) {
                    printUsage();
                    return;
                }
                else {
                    System.out.println("Error: Unknown Option!");
                    printUsage();
                    return;
                }
            }

            Document doc = formatter.read(arg);
            doc = formatter.format(doc);
            formatter.write(doc);
            System.err.println();
        }
    }

    /**
     * Describe <code>printUsage</code> method here.
     *
     */
    public static void printUsage()
    {
        System.out.println("Usage: java XmlFormatter [-option...] XMLFILE");
        System.out.println("option:");
        System.out.println("-s                 shuffle node siblings ON.");
        System.out.println("-t                 purge text nodes ON.");
        System.out.println("-a                 purge attributes ON.");
        System.out.println("-m`NNN'            set max node size to `NNN'.");
        System.out.println("-h, --help         print this help, then exit.");
    }

    /**
     * Describe <code>read</code> method here.
     *
     * @param filename a <code>String</code> value
     * @return a <code>Document</code> value
     */
    public Document read(String filename)
    {
        Document doc = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            factory.setValidating(false);
            factory.setNamespaceAware(false);

            DocumentBuilder builder = factory.newDocumentBuilder();
            doc = builder.parse(filename);

        } catch (Exception e) {
            e.printStackTrace();
        }
        return doc;
    }

    /**
     * Describe <code>format</code> method here.
     *
     * @param oldDoc a <code>Document</code> value
     * @return a <code>Document</code> value
     */
    public Document format(Document oldDoc)
    {
        Document newDoc = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            newDoc = builder.newDocument();
            traverse(newDoc, oldDoc, newDoc);

        } catch (Exception e) {
            e.printStackTrace();
        }
        return newDoc;
    }

    /**
     * Describe <code>traverse</code> method here.
     *
     * @param doc a <code>Document</code> value
     * @param oldNode a <code>Node</code> value
     * @param newNode a <code>Node</code> value
     */
    protected void traverse(Document doc, Node oldNode, Node newNode)
    {
        if (oldNode == null) return;

        switch (oldNode.getNodeType()) {
        case Node.DOCUMENT_NODE:

            Node oldRoot = ((Document)oldNode).getDocumentElement();
            Node newRoot = imports(doc, oldRoot);
            newNode.appendChild(newRoot);
            this.elementSize++;
            traverse(doc, oldRoot, newRoot);
            break;

        case Node.ELEMENT_NODE:
            NodeList oldChildren = oldNode.getChildNodes();

            if (oldChildren != null) {
                for (int i = 0; i < oldChildren.getLength(); i++) {

                    // restrict node size
                    if (this.maxElementSize != 0
                        && this.maxElementSize <= this.elementSize) break;

                    Node newChild = imports(doc, oldChildren.item(i));

                    // if TEXT_NODE
                    if (this.purgeTextNodes
                        && (newChild.getNodeType() == Node.TEXT_NODE
                            || newChild.getNodeType() == Node.CDATA_SECTION_NODE)) {
                        continue;
                    }

                    NodeList newChildren = newNode.getChildNodes();
                    int random = (int) Math.floor((newChildren.getLength()+1) * Math.random());

                    // append child
                    if (this.shuffleSiblings
                        && newChildren != null
                        && random < newChildren.getLength()) {
                        newNode.insertBefore(newChild, newChildren.item(random));
                    } else {
                        newNode.appendChild(newChild);
                    }

                    this.elementSize++;
                    traverse(doc, oldChildren.item(i), newChild);
                }
            }
            break;
        }
    }

    /**
     * Describe <code>imports</code> method here.
     *
     * @param newDoc a <code>Document</code> value
     * @param oldNode a <code>Node</code> value
     * @return a <code>Node</code> value
     */
    protected Node imports(Document newDoc, Node oldNode)
    {
        Node newNode = newDoc.importNode(oldNode, false);
        if (this.purgeAttributes && (newNode.getAttributes() != null)) {
            NamedNodeMap attrMap = newNode.getAttributes();
            for (int i = 0; i < attrMap.getLength(); i++) {
                ((Element)newNode).removeAttribute(attrMap.item(i).getNodeName());
            }
        }
        return newNode;
    }


    /**
     * Describe <code>write</code> method here.
     *
     * @param doc a <code>Document</code> value
     */
    public void write(Document doc)
    {
        try {
            // Xerces XMLSerializer
            OutputFormat formatter = new OutputFormat();
            formatter.setPreserveSpace(true);
            XMLSerializer serializer = new XMLSerializer(System.out, formatter);
            serializer.serialize(doc);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

} // XmlFormatter
